import gc
import os

import numpy as np
import sklearn.metrics
import torch
from torch.utils.data import DataLoader

import models
import utils


class Trainer(object):

    def __init__(self, args):
        # mkdir path
        if not os.path.exists('{}/events'.format(args.exp_path)):
            os.makedirs('{}/events'.format(args.exp_path))
        if not os.path.exists('{}/images'.format(args.exp_path)):
            os.makedirs('{}/images'.format(args.exp_path))
        if not os.path.exists('{}/logs'.format(args.exp_path)):
            os.makedirs('{}/logs'.format(args.exp_path))
        if not os.path.exists('{}/checkpoints'.format(args.exp_path)):
            os.makedirs('{}/checkpoints'.format(args.exp_path))
        if not os.path.exists('{}/output'.format(args.exp_path)):
            os.makedirs('{}/output'.format(args.exp_path))

        # logger
        if args.trainer['tensorboard'] and not args.extract:
            try:
                from tensorboardX import SummaryWriter
            except:
                raise Exception("Please switch off \"tensorboard\" "
                                "in your config file if you do not "
                                "want to use it, otherwise install it.")
            self.tb_logger = SummaryWriter('{}/events'.format(
                args.exp_path))
        else:
            self.tb_logger = None

        if args.extract:
            self.logger = utils.create_logger(
                'global_logger',
                '{}/logs/log_extract.txt'.format(args.exp_path))
        else:
            self.logger = utils.create_logger(
                'global_logger',
                '{}/logs/log_train.txt'.format(args.exp_path))
            self.val_logger = utils.create_logger(
                'global_logger',
                '{}/logs/log_val.txt'.format(args.exp_path))

        # Data loader
        from dataset import ImageDataset

        if not args.extract:
            train_dataset = ImageDataset(args.data, 'train')
            print("loading training dataset")
            self.train_loader = DataLoader(train_dataset,
                                           batch_size=args.data['batch_size'],
                                           shuffle=True,
                                           num_workers=args.data['workers'],
                                           pin_memory=False)

            print("loading valid dataset")
            val_dataset = ImageDataset(args.data, 'val')
            self.val_loader = DataLoader(
                val_dataset,
                batch_size=args.data['batch_size_test'],
                shuffle=False,
                num_workers=args.data['workers'],
                pin_memory=False)
        else:
            test_dataset = ImageDataset(args.data, 'test')
            self.test_loader = DataLoader(
                test_dataset,
                batch_size=args.data['batch_size_test'],
                shuffle=False,
                num_workers=args.data['workers'],
                pin_memory=False)

        # create model
        if not args.extract:
            self.model = models.__dict__[args.model['arch']](args.model,
                                                             dist_model=False, alpha=self.train_loader.dataset.alpha)
        else:
            self.model = models.__dict__[args.model['arch']](args.model, dist_model=False, alpha=1)

        # optionally resume from a checkpoint
        # assert not (args.load_iter is not None and args.load_path is not None)
        if args.load_iter is not None:
            self.model.load_state("{}/checkpoints".format(args.exp_path),
                                  args.load_iter, args.resume)
            self.start_iter = args.load_iter
        else:
            self.start_iter = 0

        if args.load_path is not None:
            self.model.load_pretrain(args.load_path)
        self.curr_step = self.start_iter

        # lr scheduler
        if not args.extract:  # train
            self.lr_scheduler = utils.StepLRScheduler(
                self.model.optim,
                args.trainer['lr_steps'],
                args.trainer['lr_mults'],
                args.trainer['lr'],
                args.trainer['warmup_lr'],
                args.trainer['warmup_steps'],
                last_iter=self.start_iter - 1)

        if not args.iou:
            self.max_iou = .0
        else:
            self.max_iou = args.iou
        self.args = args

    def run(self):
        if not self.args.extract:
            self.train()
        else:
            self.extract()
        return

    def train(self):

        recorder = {}
        for rec in self.args.trainer['loss_record']:
            recorder[rec] = utils.AverageMeter(10)

        self.model.switch_to('train')
        self.model.stepsize = self.train_loader.dataset.step
        for epoch in range(self.args.trainer["epoch"]):
            for i, (img, voxel) in enumerate(self.train_loader):
                self.curr_step = self.curr_step + 1
                self.lr_scheduler.step(self.curr_step)
                curr_lr = self.lr_scheduler.get_lr()[0]

                assert img.shape[0] > 0
                img = img.cuda()
                voxel = voxel.cuda()
                voxel_bin = voxel.clone()
                voxel_bin[voxel_bin > 0] = 1
                self.model.set_input(img, (voxel, voxel_bin))
                output = self.model.model(self.model.input)
                loss_dict = self.model.step(output)
                for k in loss_dict.keys():
                    recorder[k].update(loss_dict[k].item())

                # logging
                if self.curr_step % self.args.trainer['print_freq'] == 0:
                    loss_str = ""
                    if self.tb_logger is not None:
                        self.tb_logger.add_scalar('lr', curr_lr, self.curr_step)
                    for k in recorder.keys():
                        if self.tb_logger is not None:
                            self.tb_logger.add_scalar('train_{}'.format(k),
                                                      recorder[k].avg,
                                                      self.curr_step + 1)
                        loss_str += '{}: {loss.val:.4g} ({loss.avg:.4g})\t'.format(
                            k, loss=recorder[k])

                    self.logger.info(
                        'Iter: [{0}]\t'.format(self.curr_step) +
                        'lr: [{0:.8f}]\t'.format(curr_lr) +
                        loss_str)
                    gc.collect()

                # validate
                if (self.curr_step + 1) % self.args.trainer['val_freq'] == 0:
                    self.validate()
                    # self.logger.info('saving parms......')
                    # self.model.save_state(
                    #     "{}/checkpoints".format(self.args.exp_path),
                    #     self.curr_step + 1)

                # save
                if (self.curr_step + 1) % self.args.trainer['save_freq'] == 0:
                    self.logger.info('saving parms......')
                    self.model.save_state(
                        "{}/checkpoints".format(self.args.exp_path),
                        self.curr_step + 1)

                # just for test
                # if (self.curr_step + 1) % 100 == 0:
                #     output = torch.argmax(output, dim=1)
                #     output = output[0].cpu().detach().numpy()
                #     utils.showVoxel(output, path='output/test{}.png'.format(self.curr_step+1))

    def validate(self):
        recorder = {}
        for rec in self.args.trainer['loss_record']:
            recorder[rec] = utils.AverageMeter(10)

        self.model.switch_to('eval')
        iou_sum = .0
        for i, (img, voxel) in enumerate(self.val_loader):
            with torch.no_grad():
                # measure data loading time
                img = img.cuda()
                voxel = voxel.cuda()
                self.model.set_input(img, voxel)
                output = self.model.model(self.model.input)
                # output = self.model.model(self.model.input)
                output = torch.softmax(output, dim=1)
                output = torch.argmax(output, dim=1)

                # calculate IoU
                iou = (torch.sum(output == voxel)).item() / (32768)
                iou_sum += iou

        iou_sum = iou_sum / self.val_loader.dataset.num
        self.val_logger.info('iou: {}'.format(iou_sum))
        if self.max_iou < iou_sum:
            self.val_logger.info(
                'Validation Iter: [{0}]\t got a better IoU: [{1}]'.format(self.curr_step + 1, iou_sum))
            self.max_iou = iou_sum
            self.model.save_state("{}/checkpoints".format(self.args.exp_path), self.curr_step + 1)

        # logging
        self.val_logger.info('Validation Iter: [{0}]\t IoU: [{1}]\t'.format(self.curr_step + 1, iou_sum))

        self.model.switch_to("train")

    def extract(self):
        self.model.switch_to('eval')
        acc_sum = .0
        iou_sum = .0
        map_sum = .0
        batch_num = 0
        idx = 0
        # part_num = len(self.test_loader.dataset.class_id)
        p_iou = np.zeros(10)
        p_count = np.zeros(10)
        for _, (img, voxel) in enumerate(self.test_loader):
            assert img.shape[0] > 0
            with torch.no_grad():
                # measure data loading time
                img = img.cuda()
                voxel = voxel.cuda()
                self.model.set_input(img)
                output = self.model.model(self.model.input)
                output = torch.softmax(output, 1)
                output_argmax = torch.argmax(output, 1)
                for k in range(output.shape[0]):
                    # print(output[k].cpu().detach().numpy().max())
                    utils.showVoxel(output_argmax[k].cpu().detach().numpy(), path=
                    "{}/output/{}.png".format(self.args.exp_path, idx))
                    utils.showVoxel(voxel[k].cpu().detach().numpy(), path=
                    "{}/output/{}_gt.png".format(self.args.exp_path, idx))
                    idx = idx + 1

                # calculate acc
                acc = (torch.sum(output_argmax == voxel)).item() / 32768
                acc_sum += acc

                # calculate part iou
                for p in range(1, 10):
                    part_vox = torch.zeros_like(voxel)
                    part_out = torch.zeros_like(output_argmax)
                    part_vox[voxel == p] = 1
                    part_out[output_argmax == p] = 1
                    itersection = torch.sum(part_out * part_vox, dim=(1, 2, 3)).long()
                    union = torch.sum(torch.ge(part_out + part_vox, 1), dim=(1, 2, 3)).long()
                    batch_iou = itersection.to(torch.float64) / (union.to(torch.float64) + 0.000001)
                    iou = torch.sum(batch_iou).item()
                    p_count[p] += torch.sum(torch.gt(batch_iou, 0)).item()
                    p_iou[p] += iou

                voxel[voxel > 0] = 1
                # calculate whole map
                batch_p = 0
                for m in range(output.shape[0]):
                    vox_m = torch.flatten(voxel[m]).cpu().numpy()
                    out_m = torch.flatten(1 - output[m, 0]).cpu().numpy()
                    precision = sklearn.metrics.average_precision_score(vox_m, out_m)
                    batch_p += precision
                batch_p = batch_p / output.shape[0]
                map_sum += batch_p
                batch_num += 1

                output_argmax[output_argmax > 0] = 1
                output_argmax = output_argmax.long()
                # calculate whole iou
                itersection = torch.sum(output_argmax * voxel, dim=(1, 2, 3)).long()
                union = torch.sum(torch.ge(output_argmax + voxel, 1), dim=(1, 2, 3)).long()
                iou = torch.sum(itersection.to(torch.float64) / (union.to(torch.float64) + 0.000001)).item()
                iou_sum += iou

        acc_sum = acc_sum / self.test_loader.dataset.num
        iou_sum = iou_sum / self.test_loader.dataset.num
        map_sum = map_sum / batch_num
        # print(p_iou, p_count)
        p_sum = p_iou / (p_count + 0.00001)
        print('acc', acc_sum, 'iou', iou_sum, 'map', map_sum)
        print(p_sum)
