import numpy as np
from utils.binvox_rw import read_as_3d_array

import matplotlib.pyplot as plt
from itertools import product
# This import registers the 3D projection, but is otherwise unused.
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

color_map = {
    1: [255, 0, 0],  # RED
    2: [0, 255, 0],  # GREEN
    3: [0, 0, 255],  # BLUE
    4: [255, 255, 0],  # YELLO
    5: [255, 0, 255],  # PURPLE
    6: [0, 255, 255],  # SKY BLUE
    7: [128, 0, 0],
    8: [0, 128, 0],
    9: [0, 0, 128],
    10: [128, 255, 0],
    11: [128, 0, 255],
    12: [0, 128, 255],
    13: [255, 128, 0],
    14: [255, 0, 128],
    15: [0, 255, 128],
    16: [178, 178, 178],
    17: [178, 0, 0],
    18: [0, 178, 0],
    19: [0, 0, 178],
}


# v_list = [0, 128, 178, 255]
# idx = 1
# for color in product(v_list, reversed(v_list), [128, 255, 0, 178]):
#     color_map[idx] = color
#     idx = idx + 1


def showVoxel(voxel, quan_value=None, path=None, display=False):
    if len(voxel.shape) == 4:
        voxel = voxel.squeeze(0)
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    ax.grid(False)
    ax.set_axis_off()
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks([])

    if quan_value:
        vox = (voxel + quan_value / 2) // quan_value
    else:
        vox = voxel

    vox_shape = list(voxel.shape)
    vox_shape.append(3)
    vox_color = np.zeros(tuple(vox_shape))
    for i in range(vox.shape[0]):
        for j in range(vox.shape[1]):
            for k in range(vox.shape[2]):
                vox_color[i, j, k] = color_map.get(int(vox[i, j, k]), [0, 0, 0])
    vox[vox > 0] = 1
    ax.voxels(vox.transpose((2, 0, 1)), facecolors=vox_color.transpose((2, 0, 1, 3)) / 255, edgecolor="k")
    if display:
        plt.show()
    if path is not None:
        plt.savefig(path, dpi=200, bbox_inches='tight')
    plt.close(fig)


def evaluate_voxel_prediction(preds, gt, thresh):
    preds_occupy = preds[:, 1, :, :] >= thresh
    diff = np.sum(np.logical_xor(preds_occupy, gt[:, 1, :, :]))
    intersection = np.sum(np.logical_and(preds_occupy, gt[:, 1, :, :]))
    union = np.sum(np.logical_or(preds_occupy, gt[:, 1, :, :]))
    num_fp = np.sum(np.logical_and(preds_occupy, gt[:, 0, :, :]))  # false positive
    num_fn = np.sum(np.logical_and(np.logical_not(preds_occupy), gt[:, 1, :, :]))  # false negative
    return np.array([diff, intersection, union, num_fp, num_fn])


def voxel2mesh(voxels, surface_view, threshold=0.5):
    cube_verts = [[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 1], [1, 0, 0], [1, 0, 1], [1, 1, 0],
                  [1, 1, 1]]  # 8 points

    cube_faces = [[0, 1, 2], [1, 3, 2], [2, 3, 6], [3, 7, 6], [0, 2, 6], [0, 6, 4], [0, 5, 1],
                  [0, 4, 5], [6, 7, 5], [6, 5, 4], [1, 7, 3], [1, 5, 7]]  # 12 face

    cube_verts = np.array(cube_verts)
    cube_faces = np.array(cube_faces) + 1

    scale = 0.01
    cube_dist_scale = 1.1
    verts = []
    faces = []
    curr_vert = 0

    positions = np.where(voxels > threshold)
    voxels[positions] = 1
    for i, j, k in zip(*positions):
        # identifies if current voxel has an exposed face
        if not surface_view or np.sum(voxels[i - 1:i + 2, j - 1:j + 2, k - 1:k + 2]) < 27:
            verts.extend(scale * (cube_verts + cube_dist_scale * np.array([[i, j, k]])))
            faces.extend(cube_faces + curr_vert)
            curr_vert += len(cube_verts)

    return np.array(verts), np.array(faces)


def write_obj(filename, verts, faces):
    """ write the verts and faces on file."""
    with open(filename, 'w') as f:
        # write vertices
        f.write('g\n# %d vertex\n' % len(verts))
        for vert in verts:
            f.write('v %f %f %f\n' % tuple(vert))

        # write faces
        f.write('# %d faces\n' % len(faces))
        for face in faces:
            f.write('f %d %d %d\n' % tuple(face))


def voxel2obj(filename, pred, threshold, surface_view=True):
    verts, faces = voxel2mesh(pred, surface_view, threshold)
    write_obj(filename, verts, faces)


if __name__ == '__main__':
    vox = np.random.rand(10, 10, 10)
    showVoxel(vox, 0.2)
